/* (c) Copyright Taiichi Yuasa and Masami Hagiya, 1984.  All rights reserved.
Copying of this file is authorized to users who have executed the true and
proper "License Agreement for Kyoto Common LISP" with SIGLISP.
*/

#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>
#include <syms.h>

#define MAXPATHLEN 256
#define FASLINK

static int
linkload (faslfile, ldargstring)
  object faslfile, ldargstring;
{
   struct filehdr fileheader;
   struct scnhdr sectionheader;
   int loadsize = 0;

   object memory, data, tempfile;
   FILE *fp;
   char filename[MAXPATHLEN], ldargstr[MAXPATHLEN], ifilename[MAXPATHLEN];
   char tempfilename[32];
   char command[MAXPATHLEN * 2];
   int i;
   char buf[BUFSIZ];
   object *old_vs_base = vs_base;
   object *old_vs_top = vs_top;

   /*
    *  convert the arguments to tangable objects (C strings)
    */
   coerce_to_filename (faslfile, filename);
   if (ldargstring)
       coerce_to_filename (ldargstring, ldargstr);
   else
       ldargstr[0] = '\0';

   /*
    *  open the faslfile, and seek to the section headers
    */
   faslfile = open_stream (faslfile, smm_input, Cnil, Kerror);
   vs_push (faslfile);
   fp = faslfile->sm.sm_fp;

   fread (&fileheader, sizeof (fileheader), 1, fp);
   fseek (fp, fileheader.f_opthdr, 1);

   /*
    *  figure out how big it is.
    */
   for (i = 0; i < fileheader.f_nscns; i++)
    {
       fread (&sectionheader, sizeof (sectionheader), 1, fp);
       if (strcmp (sectionheader.s_name, ".text") == 0 ||
	   strcmp (sectionheader.s_name, ".data") == 0 ||
	   strcmp (sectionheader.s_name, ".bss") == 0)
	   loadsize += sectionheader.s_size;
    }

   /*
    *  read the fasl vector
    */
   fseek(fp, fileheader.f_symptr + (SYMESZ * fileheader.f_nsyms), 0);
   fread(&i, sizeof (i), 1, fp);
   fseek(fp, i - sizeof (i), 1);
   while ((i = getc (fp)) == 0)
    ;
   ungetc (i, fp);

   data = read_fasl_vector (faslfile);
   vs_push (data);

   /*
    *  we're done with the fasl file.
    */
   close_stream (faslfile, TRUE);

AGAIN:
   /*
    *  snarf loadsize bytes of memory.
    */
   memory = alloc_object (t_cfdata);
   memory->cfd.cfd_self = NULL;
   memory->cfd.cfd_start = NULL;
   memory->cfd.cfd_size = loadsize;
   vs_push (memory);
   memory->cfd.cfd_start = alloc_contblock (memory->cfd.cfd_size);

   /*
    * construct a temp file name.
    */
   sprintf (tempfilename, "/tmp/fasltemp%d", getpid ());
		     
   /*
    *  create the ifile.
    * 
    */
   sprintf (ifilename, "/tmp/faslifile%d", getpid ());
   fp = fopen (ifilename, "w");
   setbuf (fp, buf);
   fprintf (fp, "\
MEMORY { user_mem : origin=0x80000, length=0x37ff000 } \n\
SECTIONS {\n\
	.phony 0x80000 (DSECT) : { \n\
		%s (.text) \n\
		. = align (65536) ; \n\
		%s (.data .bss) \n\
	} \n\
	.load 0x%x : { %s %s }\n\
} -o %s -s\n\
", kcl_self, kcl_self, memory->cfd.cfd_start, filename, ldargstr, tempfilename);
   fclose (fp);

   /*
    *  run the loader
    */
   sprintf (command, "/bin/ld %s", ifilename );
   if (system (command) != 0)
       FEerror("The linkage editor failed.", 0);

   /*
    * open the tempfile and make sire it didn't get bigger.
    */
   fp = fopen (tempfilename, "r");
   setbuf (fp, buf);

   if (fseek (fp, sizeof (struct filehdr) + sizeof (struct aouthdr), 0) < 0)
       error ("file seek error");
   do {
      fread (&sectionheader, sizeof (sectionheader), 1, fp);
   } while (strcmp (sectionheader.s_name, ".load"));

   if (sectionheader.s_size != loadsize)
    {
       insert_contblock (memory->cfd.cfd_start, memory->cfd.cfd_size);
       loadsize = sectionheader.s_size;
       goto AGAIN;
    }
   if (fseek (fp, sectionheader.s_scnptr, 0) < 0)
    error ("file seek error");
  
   fread (memory->cfd.cfd_start, loadsize, 1, fp);

   fclose (fp);

   unlink (ifilename);
   unlink (tempfilename);

   call_init(0,memory,data);

   vs_base = old_vs_base;
   vs_top = old_vs_top;

   return memory->cfd.cfd_size;
}

int
fasload(faslfile)
  object faslfile;
{
   return linkload (faslfile, NULL);
}

static void
siLfaslink()
{
	bds_ptr old_bds_top;
	int i;
	object package;

	check_arg(2);
	check_type_or_pathname_string_symbol_stream(&vs_base[0]);
	check_type_string(&vs_base[1]);
	vs_base[0] = coerce_to_pathname(vs_base[0]);
	vs_base[0]->pn.pn_type = FASL_string;
	vs_base[0] = namestring(vs_base[0]);
	package = symbol_value(Vpackage);
	old_bds_top = bds_top;
	bds_bind(Vpackage, package);
	i = linkload(vs_base[0], vs_base[1]);
	bds_unwind(old_bds_top);
	vs_top = vs_base;
	vs_push(make_fixnum(i));
}
